import os
from twisted.internet import reactor
from twisted.python import usage
from twisted.application.service import MultiService
from twisted.application.internet import (TimerService,
                                          StreamServerEndpointService)
from twisted.internet import endpoints
from twisted.internet import protocol

from autobahn.twisted.websocket import WebSocketServerFactory

from . import transit_server
from .usage import create_usage_tracker
from .increase_rlimits import increase_rlimits
from .database import get_db

LONGDESC = """\
This plugin sets up a 'Transit Relay' server for magic-wormhole. This service
listens for TCP connections, finds pairs which present the same handshake, and
glues the two TCP sockets together.
"""

class Options(usage.Options):
    synopsis = "[--port=] [--log-fd] [--blur-usage=] [--usage-db=]"
    longdesc = LONGDESC

    optParameters = [
        ("port", "p", "tcp:4001:interface=\:\:", "endpoint to listen on"),
        ("websocket", "w", None, "endpoint to listen for WebSocket connections"),
        ("websocket-url", "u", None, "WebSocket URL (derived from endpoint if not provided)"),
        ("blur-usage", None, None, "blur timestamps and data sizes in logs"),
        ("log-fd", None, None, "write JSON usage logs to this file descriptor"),
        ("usage-db", None, None, "record usage data (SQLite)"),
        ]

    def opt_blur_usage(self, arg):
        self["blur-usage"] = int(arg)


def makeService(config, reactor=reactor):
    increase_rlimits()
    tcp_ep = endpoints.serverFromString(reactor, config["port"]) # to listen
    ws_ep = (
        endpoints.serverFromString(reactor, config["websocket"])
        if config["websocket"] is not None
        else None
    )
    log_file = (
        os.fdopen(int(config["log-fd"]), "w")
        if config["log-fd"] is not None
        else None
    )
    db = None if config["usage-db"] is None else get_db(config["usage-db"])
    usage = create_usage_tracker(
        blur_usage=config["blur-usage"],
        log_file=log_file,
        usage_db=db,
    )
    transit = transit_server.Transit(usage, reactor.seconds)
    tcp_factory = protocol.ServerFactory()
    tcp_factory.protocol = transit_server.TransitConnection
    tcp_factory.log_requests = False

    if ws_ep is not None:
        ws_url = config["websocket-url"]
        if ws_url is None:
            # we're using a "private" attribute here but I don't see
            # any useful alternative unless we also want to parse
            # Twisted endpoint-strings.
            ws_url = "ws://localhost:{}/".format(ws_ep._port)
            print("Using WebSocket URL '{}'".format(ws_url))
        ws_factory = WebSocketServerFactory(ws_url)
        ws_factory.protocol = transit_server.WebSocketTransitConnection
        ws_factory.transit = transit
        ws_factory.log_requests = False

    tcp_factory.transit = transit
    parent = MultiService()
    StreamServerEndpointService(tcp_ep, tcp_factory).setServiceParent(parent)
    if ws_ep is not None:
        StreamServerEndpointService(ws_ep, ws_factory).setServiceParent(parent)
    TimerService(5*60.0, transit.update_stats).setServiceParent(parent)
    return parent



from unittest import mock
from twisted.trial import unittest
from ..increase_rlimits import increase_rlimits

class RLimits(unittest.TestCase):
    def test_rlimit(self):
        def patch_r(name, *args, **kwargs):
            return mock.patch("wormhole_transit_relay.increase_rlimits." + name, *args, **kwargs)
        fakelog = []
        def checklog(*expected):
            self.assertEqual(fakelog, list(expected))
            fakelog[:] = []
        NF = "NOFILE"
        mock_NF = patch_r("RLIMIT_NOFILE", NF)

        with patch_r("log.msg", fakelog.append):
            with patch_r("getrlimit", None):
                increase_rlimits()
            checklog("unable to import 'resource', leaving rlimit alone")

            with mock_NF:
                with patch_r("getrlimit", return_value=(20000, 30000)) as gr:
                    increase_rlimits()
                    self.assertEqual(gr.mock_calls, [mock.call(NF)])
                    checklog("RLIMIT_NOFILE.soft was 20000, leaving it alone")

                with patch_r("getrlimit", return_value=(10, 30000)) as gr:
                    with patch_r("setrlimit", side_effect=TypeError("other")):
                        with patch_r("log.err") as err:
                            increase_rlimits()
                        self.assertEqual(err.mock_calls, [mock.call()])
                        checklog("changing RLIMIT_NOFILE from (10,30000) to (30000,30000)",
                                 "other error during setrlimit, leaving it alone")

                    for maxlimit in [40000, 20000, 9000, 2000, 1000]:
                        def setrlimit(which, newlimit):
                            if newlimit[0] > maxlimit:
                                raise ValueError("nope")
                            return None
                        calls = []
                        expected = []
                        for tries in [30000, 10000, 3200, 1024]:
                            calls.append(mock.call(NF, (tries, 30000)))
                            expected.append("changing RLIMIT_NOFILE from (10,30000) to (%d,30000)" % tries)
                            if tries > maxlimit:
                                expected.append("error during setrlimit: nope")
                            else:
                                expected.append("setrlimit successful")
                                break
                        else:
                            expected.append("unable to change rlimit, leaving it alone")

                        with patch_r("setrlimit", side_effect=setrlimit) as sr:
                            increase_rlimits()
                        self.assertEqual(sr.mock_calls, calls)
                        checklog(*expected)



from twisted.application.service import ServiceMaker

TransitRelay = ServiceMaker(
    "Magic-Wormhole Transit Relay", # name
    "wormhole_transit_relay.server_tap", # module
    "Provide the Transit Relay server for Magic-Wormhole clients.", # desc
    "transitrelay", # tapname
    )



import os, io, json
from unittest import mock
from twisted.trial import unittest
from ..transit_server import Transit
from ..usage import create_usage_tracker
from .. import database

class DB(unittest.TestCase):

    def test_db(self):

        T = 1519075308.0

        class Timer:
            t = T
            def __call__(self):
                return self.t
        get_time = Timer()

        d = self.mktemp()
        os.mkdir(d)
        usage_db = os.path.join(d, "usage.sqlite")
        db = database.get_db(usage_db)
        t = Transit(
            create_usage_tracker(blur_usage=None, log_file=None, usage_db=db),
            get_time,
        )
        self.assertEqual(len(t.usage._backends), 1)
        usage = list(t.usage._backends)[0]

        get_time.t = T + 1
        usage.record_usage(started=123, mood="happy", total_bytes=100,
                           total_time=10, waiting_time=2)
        t.update_stats()

        self.assertEqual(db.execute("SELECT * FROM `usage`").fetchall(),
                         [dict(result="happy", started=123,
                               total_bytes=100, total_time=10, waiting_time=2),
                          ])
        self.assertEqual(db.execute("SELECT * FROM `current`").fetchone(),
                         dict(rebooted=T+0, updated=T+1,
                              incomplete_bytes=0,
                              waiting=0, connected=0))

        get_time.t = T + 2
        usage.record_usage(started=150, mood="errory", total_bytes=200,
                           total_time=11, waiting_time=3)
        t.update_stats()
        self.assertEqual(db.execute("SELECT * FROM `usage`").fetchall(),
                         [dict(result="happy", started=123,
                               total_bytes=100, total_time=10, waiting_time=2),
                          dict(result="errory", started=150,
                               total_bytes=200, total_time=11, waiting_time=3),
                          ])
        self.assertEqual(db.execute("SELECT * FROM `current`").fetchone(),
                         dict(rebooted=T+0, updated=T+2,
                              incomplete_bytes=0,
                              waiting=0, connected=0))

        get_time.t = T + 3
        t.update_stats()
        self.assertEqual(db.execute("SELECT * FROM `current`").fetchone(),
                         dict(rebooted=T+0, updated=T+3,
                              incomplete_bytes=0,
                              waiting=0, connected=0))

    def test_no_db(self):
        t = Transit(
            create_usage_tracker(blur_usage=None, log_file=None, usage_db=None),
            lambda: 0,
        )
        self.assertEqual(0, len(t.usage._backends))


class LogToStdout(unittest.TestCase):
    def test_log(self):
        # emit lines of JSON to log_file, if set
        log_file = io.StringIO()
        t = Transit(
            create_usage_tracker(blur_usage=None, log_file=log_file, usage_db=None),
            lambda: 0,
        )
        with mock.patch("time.time", return_value=133):
            t.usage.record(
                started=123,
                buddy_started=125,
                result="happy",
                bytes_sent=100,
                buddy_bytes=0,
            )
        self.assertEqual(json.loads(log_file.getvalue()),
                         {"started": 123, "total_time": 10,
                          "waiting_time": 2, "total_bytes": 100,
                          "mood": "happy"})

    def test_log_blurred(self):
        # if blurring is enabled, timestamps should be rounded to the
        # requested amount, and sizes should be rounded up too
        log_file = io.StringIO()
        t = Transit(
            create_usage_tracker(blur_usage=60, log_file=log_file, usage_db=None),
            lambda: 0,
        )

        with mock.patch("time.time", return_value=123 + 10):
            t.usage.record(
                started=123,
                buddy_started=125,
                result="happy",
                bytes_sent=11999,
                buddy_bytes=0,
            )
        print(log_file.getvalue())
        self.assertEqual(json.loads(log_file.getvalue()),
                         {"started": 120, "total_time": 10,
                          "waiting_time": 2, "total_bytes": 20000,
                          "mood": "happy"})

    def test_do_not_log(self):
        t = Transit(
            create_usage_tracker(blur_usage=60, log_file=None, usage_db=None),
            lambda: 0,
        )
        t.usage.record(
            started=123,
            buddy_started=124,
            result="happy",
            bytes_sent=11999,
            buddy_bytes=12,
        )



import re
import time
from twisted.python import log
from twisted.protocols.basic import LineReceiver
from autobahn.twisted.websocket import WebSocketServerProtocol


SECONDS = 1.0
MINUTE = 60*SECONDS
HOUR = 60*MINUTE
DAY = 24*HOUR
MB = 1000*1000


from wormhole_transit_relay.server_state import (
    TransitServerState,
    PendingRequests,
    ActiveConnections,
    ITransitClient,
)
from zope.interface import implementer


@implementer(ITransitClient)
class TransitConnection(LineReceiver):
    delimiter = b'\n'
    # maximum length of a line we will accept before the handshake is complete.
    # This must be >= to the longest possible handshake message.

    MAX_LENGTH = 1024
    started_time = None

    def send(self, data):
        """
        ITransitClient API
        """
        self.transport.write(data)

    def disconnect(self):
        """
        ITransitClient API
        """
        self.transport.loseConnection()

    def connect_partner(self, other):
        """
        ITransitClient API
        """
        self._buddy = other
        self._buddy._client.transport.registerProducer(self.transport, True)

    def disconnect_partner(self):
        """
        ITransitClient API
        """
        assert self._buddy is not None, "internal error: no buddy"
        if self.factory.log_requests:
            log.msg("buddy_disconnected {}".format(self._buddy.get_token()))
        self._buddy._client.disconnect()
        self._buddy = None

    def connectionMade(self):
        # ideally more like self._reactor.seconds() ... but Twisted
        # doesn't have a good way to get the reactor for a protocol
        # (besides "use the global one")
        self.started_time = time.time()
        self._state = TransitServerState(
            self.factory.transit.pending_requests,
            self.factory.transit.usage,
        )
        self._state.connection_made(self)
        self.transport.setTcpKeepAlive(True)

        # uncomment to turn on state-machine tracing
        # def tracer(oldstate, theinput, newstate):
        #     print("TRACE: {}: {} --{}--> {}".format(id(self), oldstate, theinput, newstate))
        # self._state.set_trace_function(tracer)

    def lineReceived(self, line):
        """
        LineReceiver API
        """
        # old: "please relay {64}\n"
        token = None
        old = re.search(br"^please relay (\w{64})$", line)
        if old:
            token = old.group(1)
            self._state.please_relay(token)

        # new: "please relay {64} for side {16}\n"
        new = re.search(br"^please relay (\w{64}) for side (\w{16})$", line)
        if new:
            token = new.group(1)
            side = new.group(2)
            self._state.please_relay_for_side(token, side)

        if token is None:
            self._state.bad_token()
        else:
            self.setRawMode()

    def rawDataReceived(self, data):
        """
        LineReceiver API
        """
        # We are an IPushProducer to our buddy's IConsumer, so they'll
        # throttle us (by calling pauseProducing()) when their outbound
        # buffer is full (e.g. when their downstream pipe is full). In
        # practice, this buffers about 10MB per connection, after which
        # point the sender will only transmit data as fast as the
        # receiver can handle it.
        self._state.got_bytes(data)

    def connectionLost(self, reason):
        self._state.connection_lost()


class Transit(object):
    """
    I manage pairs of simultaneous connections to a secondary TCP port,
    both forwarded to the other. Clients must begin each connection with
    "please relay TOKEN for SIDE\n" (or a legacy form without the "for
    SIDE"). Two connections match if they use the same TOKEN and have
    different SIDEs (the redundant connections are dropped when a match is
    made). Legacy connections match any with the same TOKEN, ignoring SIDE
    (so two legacy connections will match each other).

    I will send "ok\n" when the matching connection is established, or
    disconnect if no matching connection is made within MAX_WAIT_TIME
    seconds. I will disconnect if you send data before the "ok\n". All data
    you get after the "ok\n" will be from the other side. You will not
    receive "ok\n" until the other side has also connected and submitted a
    matching token (and differing SIDE).

    In addition, the connections will be dropped after MAXLENGTH bytes have
    been sent by either side, or MAXTIME seconds have elapsed after the
    matching connections were established. A future API will reveal these
    limits to clients instead of causing mysterious spontaneous failures.

    These relay connections are not half-closeable (unlike full TCP
    connections, applications will not receive any data after half-closing
    their outgoing side). Applications must negotiate shutdown with their
    peer and not close the connection until all data has finished
    transferring in both directions. Applications which only need to send
    data in one direction can use close() as usual.
    """

    # TODO: unused
    MAX_WAIT_TIME = 30*SECONDS
    # TODO: unused
    MAXLENGTH = 10*MB
    # TODO: unused
    MAXTIME = 60*SECONDS

    def __init__(self, usage, get_timestamp):
        self.active_connections = ActiveConnections()
        self.pending_requests = PendingRequests(self.active_connections)
        self.usage = usage
        self._timestamp = get_timestamp
        self._rebooted = self._timestamp()

    def update_stats(self):
        # TODO: when a connection is half-closed, len(active) will be odd. a
        # moment later (hopefully) the other side will disconnect, but
        # _update_stats isn't updated until later.

        # "waiting" doesn't count multiple parallel connections from the same
        # side
        self.usage.update_stats(
            rebooted=self._rebooted,
            updated=self._timestamp(),
            connected=len(self.active_connections._connections),
            waiting=len(self.pending_requests._requests),
            incomplete_bytes=sum(
                tc._total_sent
                for tc in self.active_connections._connections
            ),
        )


@implementer(ITransitClient)
class WebSocketTransitConnection(WebSocketServerProtocol):
    started_time = None

    def send(self, data):
        """
        ITransitClient API
        """
        self.sendMessage(data, isBinary=True)

    def disconnect(self):
        """
        ITransitClient API
        """
        self.sendClose(1000, None)

    def connect_partner(self, other):
        """
        ITransitClient API
        """
        self._buddy = other
        self._buddy._client.transport.registerProducer(self.transport, True)

    def disconnect_partner(self):
        """
        ITransitClient API
        """
        assert self._buddy is not None, "internal error: no buddy"
        if self.factory.log_requests:
            log.msg("buddy_disconnected {}".format(self._buddy.get_token()))
        self._buddy._client.disconnect()
        self._buddy = None

    def connectionMade(self):
        """
        IProtocol API
        """
        super(WebSocketTransitConnection, self).connectionMade()
        self.started_time = time.time()
        self._first_message = True
        self._state = TransitServerState(
            self.factory.transit.pending_requests,
            self.factory.transit.usage,
        )

        # uncomment to turn on state-machine tracing
        # def tracer(oldstate, theinput, newstate):
        #    print("WSTRACE: {}: {} --{}--> {}".format(id(self), oldstate, theinput, newstate))
        # self._state.set_trace_function(tracer)

    def onOpen(self):
        self._state.connection_made(self)

    def onMessage(self, payload, isBinary):
        """
        We may have a 'handshake' on our hands or we may just have some bytes to relay
        """
        if not isBinary:
            raise ValueError(
                "All messages must be binary"
            )
        if self._first_message:
            self._first_message = False
            token = None
            old = re.search(br"^please relay (\w{64})$", payload)
            if old:
                token = old.group(1)
                self._state.please_relay(token)

            # new: "please relay {64} for side {16}\n"
            new = re.search(br"^please relay (\w{64}) for side (\w{16})$", payload)
            if new:
                token = new.group(1)
                side = new.group(2)
                self._state.please_relay_for_side(token, side)

            if token is None:
                self._state.bad_token()
        else:
            self._state.got_bytes(payload)

    def onClose(self, wasClean, code, reason):
        """
        IWebSocketChannel API
        """
        self._state.connection_lost()



from collections import defaultdict

import automat
from twisted.python import log
from zope.interface import (
    Interface,
    Attribute,
)


class ITransitClient(Interface):
    """
    Represents the client side of a connection to this transit
    relay. This is used by TransitServerState instances.
    """

    started_time = Attribute("timestamp when the connection was established")

    def send(data):
        """
        Send some byets to the client
        """

    def disconnect():
        """
        Disconnect the client transport
        """

    def connect_partner(other):
        """
        Hook up to our partner.
        :param ITransitClient other: our partner
        """

    def disconnect_partner():
        """
        Disconnect our partner's transport
        """


class ActiveConnections(object):
    """
    Tracks active connections.

    A connection is 'active' when both sides have shown up and they
    are glued together (and thus could be passing data back and forth
    if any is flowing).
    """
    def __init__(self):
        self._connections = set()

    def register(self, side0, side1):
        """
        A connection has become active so register both its sides

        :param TransitConnection side0: one side of the connection
        :param TransitConnection side1: one side of the connection
        """
        self._connections.add(side0)
        self._connections.add(side1)

    def unregister(self, side):
        """
        One side of a connection has become inactive.

        :param TransitConnection side: an inactive side of a connection
        """
        self._connections.discard(side)


class PendingRequests(object):
    """
    Tracks outstanding (non-"active") requests.

    We register client connections against the tokens we have
    received. When the other side shows up we can thus match it to the
    correct partner connection. At this point, the connection becomes
    "active" is and is thus no longer "pending" and so will no longer
    be in this collection.
    """

    def __init__(self, active_connections):
        """
        :param active_connections: an instance of ActiveConnections where
            connections are put when both sides arrive.
        """
        self._requests = defaultdict(set) # token -> set((side, TransitConnection))
        self._active = active_connections

    def unregister(self, token, side, tc):
        """
        We no longer care about a particular client (e.g. it has
        disconnected).
        """
        if token in self._requests:
            self._requests[token].discard((side, tc))
            if not self._requests[token]:
                # no more sides; token is dead
                del self._requests[token]
        self._active.unregister(tc)

    def register(self, token, new_side, new_tc):
        """
        A client has connected and successfully offered a token (and
        optional 'side' token). If this is the first one for this
        token, we merely remember it. If it is the second side for
        this token we connect them together.

        :param bytes token: the token for this connection.

        :param bytes new_side: None or the side token for this connection

        :param TransitServerState new_tc: the state-machine of the connection

        :returns bool: True if we are the first side to register this
            token
        """
        potentials = self._requests[token]
        for old in potentials:
            (old_side, old_tc) = old
            if ((old_side is None)
                or (new_side is None)
                or (old_side != new_side)):
                # we found a match

                # drop and stop tracking the rest
                potentials.remove(old)
                for (_, leftover_tc) in potentials.copy():
                    # Don't record this as errory. It's just a spare connection
                    # from the same side as a connection that got used. This
                    # can happen if the connection hint contains multiple
                    # addresses (we don't currently support those, but it'd
                    # probably be useful in the future).
                    leftover_tc.partner_connection_lost()
                self._requests.pop(token, None)

                # glue the two ends together
                self._active.register(new_tc, old_tc)
                new_tc.got_partner(old_tc)
                old_tc.got_partner(new_tc)
                return False

        potentials.add((new_side, new_tc))
        return True
        # TODO: timer


class TransitServerState(object):
    """
    Encapsulates the state-machine of the server side of a transit
    relay connection.

    Once the protocol has been told to relay (or to relay for a side)
    it starts passing all received bytes to the other side until it
    closes.
    """

    _machine = automat.MethodicalMachine()
    _client = None
    _buddy = None
    _token = None
    _side = None
    _first = None
    _mood = "empty"
    _total_sent = 0

    def __init__(self, pending_requests, usage_recorder):
        self._pending_requests = pending_requests
        self._usage = usage_recorder

    def get_token(self):
        """
        :returns str: a string describing our token. This will be "-" if
            we have no token yet, or "{16 chars}-<unsided>" if we have
            just a token or "{16 chars}-{16 chars}" if we have a token and
            a side.
        """
        d = "-"
        if self._token is not None:
            d = self._token[:16].decode("ascii")

            if self._side is not None:
                d += "-" + self._side.decode("ascii")
            else:
                d += "-<unsided>"
        return d

    @_machine.input()
    def connection_made(self, client):
        """
        A client has connected. May only be called once.

        :param ITransitClient client: our client.
        """
        # NB: the "only called once" is enforced by the state-machine;
        # this input is only valid for the "listening" state, to which
        # we never return.

    @_machine.input()
    def please_relay(self, token):
        """
        A 'please relay X' message has been received (the original version
        of the protocol).
        """

    @_machine.input()
    def please_relay_for_side(self, token, side):
        """
        A 'please relay X for side Y' message has been received (the
        second version of the protocol).
        """

    @_machine.input()
    def bad_token(self):
        """
        A bad token / relay line was received (e.g. couldn't be parsed)
        """

    @_machine.input()
    def got_partner(self, client):
        """
        The partner for this relay session has been found
        """

    @_machine.input()
    def connection_lost(self):
        """
        Our transport has failed.
        """

    @_machine.input()
    def partner_connection_lost(self):
        """
        Our partner's transport has failed.
        """

    @_machine.input()
    def got_bytes(self, data):
        """
        Some bytes have arrived (that aren't part of the handshake)
        """

    @_machine.output()
    def _remember_client(self, client):
        self._client = client

    # note that there is no corresponding "_forget_client" because we
    # may still want to access it after it is gone .. for example, to
    # get the .started_time for logging purposes

    @_machine.output()
    def _register_token(self, token):
        return self._real_register_token_for_side(token, None)

    @_machine.output()
    def _register_token_for_side(self, token, side):
        return self._real_register_token_for_side(token, side)

    @_machine.output()
    def _unregister(self):
        """
        remove us from the thing that remembers tokens and sides
        """
        return self._pending_requests.unregister(self._token, self._side, self)

    @_machine.output()
    def _send_bad(self):
        self._mood = "errory"
        self._client.send(b"bad handshake\n")
        if self._client.factory.log_requests:
            log.msg("transit handshake failure")

    @_machine.output()
    def _send_ok(self):
        self._client.send(b"ok\n")

    @_machine.output()
    def _send_impatient(self):
        self._client.send(b"impatient\n")
        if self._client.factory.log_requests:
            log.msg("transit impatience failure")

    @_machine.output()
    def _count_bytes(self, data):
        self._total_sent += len(data)

    @_machine.output()
    def _send_to_partner(self, data):
        self._buddy._client.send(data)

    @_machine.output()
    def _connect_partner(self, client):
        self._buddy = client
        self._client.connect_partner(client)

    @_machine.output()
    def _disconnect(self):
        self._client.disconnect()

    @_machine.output()
    def _disconnect_partner(self):
        self._client.disconnect_partner()

    # some outputs to record "usage" information ..
    @_machine.output()
    def _record_usage(self):
        if self._mood == "jilted":
            if self._buddy and self._buddy._mood == "happy":
                return
        self._usage.record(
            started=self._client.started_time,
            buddy_started=self._buddy._client.started_time if self._buddy is not None else None,
            result=self._mood,
            bytes_sent=self._total_sent,
            buddy_bytes=self._buddy._total_sent if self._buddy is not None else None
        )

    # some outputs to record the "mood" ..
    @_machine.output()
    def _mood_happy(self):
        self._mood = "happy"

    @_machine.output()
    def _mood_lonely(self):
        self._mood = "lonely"

    @_machine.output()
    def _mood_redundant(self):
        self._mood = "redundant"

    @_machine.output()
    def _mood_impatient(self):
        self._mood = "impatient"

    @_machine.output()
    def _mood_errory(self):
        self._mood = "errory"

    @_machine.output()
    def _mood_happy_if_first(self):
        """
        We disconnected first so we're only happy if we also connected
        first.
        """
        if self._first:
            self._mood = "happy"
        else:
            self._mood = "jilted"

    def _real_register_token_for_side(self, token, side):
        """
        A client has connected and sent a valid version 1 or version 2
        handshake. If the former, `side` will be None.

        In either case, we remember the tokens and register
        ourselves. This might result in 'got_partner' notifications to
        two state-machines if this is the second side for a given token.

        :param bytes token: the token
        :param bytes side: The side token (or None)
        """
        self._token = token
        self._side = side
        self._first = self._pending_requests.register(token, side, self)

    @_machine.state(initial=True)
    def listening(self):
        """
        Initial state, awaiting connection.
        """

    @_machine.state()
    def wait_relay(self):
        """
        Waiting for a 'relay' message
        """

    @_machine.state()
    def wait_partner(self):
        """
        Waiting for our partner to connect
        """

    @_machine.state()
    def relaying(self):
        """
        Relaying bytes to our partner
        """

    @_machine.state()
    def done(self):
        """
        Terminal state
        """

    listening.upon(
        connection_made,
        enter=wait_relay,
        outputs=[_remember_client],
    )
    listening.upon(
        connection_lost,
        enter=done,
        outputs=[_mood_errory],
    )

    wait_relay.upon(
        please_relay,
        enter=wait_partner,
        outputs=[_mood_lonely, _register_token],
    )
    wait_relay.upon(
        please_relay_for_side,
        enter=wait_partner,
        outputs=[_mood_lonely, _register_token_for_side],
    )
    wait_relay.upon(
        bad_token,
        enter=done,
        outputs=[_mood_errory, _send_bad, _disconnect, _record_usage],
    )
    wait_relay.upon(
        got_bytes,
        enter=done,
        outputs=[_count_bytes, _mood_errory, _disconnect, _record_usage],
    )
    wait_relay.upon(
        connection_lost,
        enter=done,
        outputs=[_disconnect, _record_usage],
    )

    wait_partner.upon(
        got_partner,
        enter=relaying,
        outputs=[_mood_happy, _send_ok, _connect_partner],
    )
    wait_partner.upon(
        connection_lost,
        enter=done,
        outputs=[_mood_lonely, _unregister, _record_usage],
    )
    wait_partner.upon(
        got_bytes,
        enter=done,
        outputs=[_mood_impatient, _send_impatient, _disconnect, _unregister, _record_usage],
    )
    wait_partner.upon(
        partner_connection_lost,
        enter=done,
        outputs=[_mood_redundant, _disconnect, _record_usage],
    )

    relaying.upon(
        got_bytes,
        enter=relaying,
        outputs=[_count_bytes, _send_to_partner],
    )
    relaying.upon(
        connection_lost,
        enter=done,
        outputs=[_mood_happy_if_first, _disconnect_partner, _unregister, _record_usage],
    )

    done.upon(
        connection_lost,
        enter=done,
        outputs=[],
    )
    done.upon(
        partner_connection_lost,
        enter=done,
        outputs=[],
    )

    # uncomment to turn on state-machine tracing
    # set_trace_function = _machine._setTrace



from twisted.trial import unittest
from .. import server_tap

PORT = "tcp:4001:interface=\:\:"

class Config(unittest.TestCase):
    def test_defaults(self):
        o = server_tap.Options()
        o.parseOptions([])
        self.assertEqual(o, {"blur-usage": None, "log-fd": None,
                             "usage-db": None, "port": PORT,
                             "websocket": None, "websocket-url": None})
    def test_blur(self):
        o = server_tap.Options()
        o.parseOptions(["--blur-usage=60"])
        self.assertEqual(o, {"blur-usage": 60, "log-fd": None,
                             "usage-db": None, "port": PORT,
                             "websocket": None, "websocket-url": None})

    def test_websocket(self):
        o = server_tap.Options()
        o.parseOptions(["--websocket=tcp:4004"])
        self.assertEqual(o, {"blur-usage": None, "log-fd": None,
                             "usage-db": None, "port": PORT,
                             "websocket": "tcp:4004", "websocket-url": None})

    def test_websocket_url(self):
        o = server_tap.Options()
        o.parseOptions(["--websocket=tcp:4004", "--websocket-url=ws://example.com/"])
        self.assertEqual(o, {"blur-usage": None, "log-fd": None,
                             "usage-db": None, "port": PORT,
                             "websocket": "tcp:4004",
                             "websocket-url": "ws://example.com/"})

    def test_string(self):
        o = server_tap.Options()
        s = str(o)
        self.assertIn("This plugin sets up a 'Transit Relay'", s)
        self.assertIn("--blur-usage=", s)
        self.assertIn("blur timestamps and data sizes in logs", s)




import os
import sqlite3
import tempfile
from pkg_resources import resource_string
from twisted.python import log

class DBError(Exception):
    pass

def get_schema(version):
    schema_bytes = resource_string("wormhole_transit_relay",
                                   "db-schemas/v%d.sql" % version)
    return schema_bytes.decode("utf-8")

## def get_upgrader(new_version):
##     schema_bytes = resource_string("wormhole_transit_relay",
##                                    "db-schemas/upgrade-to-v%d.sql" % new_version)
##     return schema_bytes.decode("utf-8")

TARGET_VERSION = 1

def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d

def _initialize_db_schema(db, target_version):
    """Creates the application schema in the given database.
    """
    log.msg("populating new database with schema v%s" % target_version)
    schema = get_schema(target_version)
    db.executescript(schema)
    db.execute("INSERT INTO version (version) VALUES (?)",
               (target_version,))
    db.commit()

def _initialize_db_connection(db):
    """Sets up the db connection object with a row factory and with necessary
    foreign key settings.
    """
    db.row_factory = dict_factory
    db.execute("PRAGMA foreign_keys = ON")
    problems = db.execute("PRAGMA foreign_key_check").fetchall()
    if problems:
        raise DBError("failed foreign key check: %s" % (problems,))

def _open_db_connection(dbfile):
    """Open a new connection to the SQLite3 database at the given path.
    """
    try:
        db = sqlite3.connect(dbfile)
        _initialize_db_connection(db)
    except (EnvironmentError, sqlite3.OperationalError, sqlite3.DatabaseError) as e:
        # this indicates that the file is not a compatible database format.
        # Perhaps it was created with an old version, or it might be junk.
        raise DBError("Unable to create/open db file %s: %s" % (dbfile, e))
    return db

def _get_temporary_dbfile(dbfile):
    """Get a temporary filename near the given path.
    """
    fd, name = tempfile.mkstemp(
        prefix=os.path.basename(dbfile) + ".",
        dir=os.path.dirname(dbfile)
    )
    os.close(fd)
    return name

def _atomic_create_and_initialize_db(dbfile, target_version):
    """Create and return a new database, initialized with the application
    schema.

    If anything goes wrong, nothing is left at the ``dbfile`` path.
    """
    temp_dbfile = _get_temporary_dbfile(dbfile)
    db = _open_db_connection(temp_dbfile)
    _initialize_db_schema(db, target_version)
    db.close()
    os.rename(temp_dbfile, dbfile)
    return _open_db_connection(dbfile)

def get_db(dbfile, target_version=TARGET_VERSION):
    """Open or create the given db file. The parent directory must exist.
    Returns the db connection object, or raises DBError.
    """
    if dbfile == ":memory:":
        db = _open_db_connection(dbfile)
        _initialize_db_schema(db, target_version)
    elif os.path.exists(dbfile):
        db = _open_db_connection(dbfile)
    else:
        db = _atomic_create_and_initialize_db(dbfile, target_version)

    version = db.execute("SELECT version FROM version").fetchone()["version"]

    ## while version < target_version:
    ##     log.msg(" need to upgrade from %s to %s" % (version, target_version))
    ##     try:
    ##         upgrader = get_upgrader(version+1)
    ##     except ValueError: # ResourceError??
    ##         log.msg(" unable to upgrade %s to %s" % (version, version+1))
    ##         raise DBError("Unable to upgrade %s to version %s, left at %s"
    ##                       % (dbfile, version+1, version))
    ##     log.msg(" executing upgrader v%s->v%s" % (version, version+1))
    ##     db.executescript(upgrader)
    ##     db.commit()
    ##     version = version+1

    if version != target_version:
        raise DBError("Unable to handle db version %s" % version)

    return db

class DBDoesntExist(Exception):
    pass

def open_existing_db(dbfile):
    assert dbfile != ":memory:"
    if not os.path.exists(dbfile):
        raise DBDoesntExist()
    return _open_db_connection(dbfile)

class DBAlreadyExists(Exception):
    pass

def create_db(dbfile):
    """Create the given db file. Refuse to touch a pre-existing file.

    This is meant for use by migration tools, to create the output target"""

    if dbfile == ":memory:":
        db = _open_db_connection(dbfile)
        _initialize_db_schema(db, TARGET_VERSION)
    elif os.path.exists(dbfile):
        raise DBAlreadyExists()
    else:
        db = _atomic_create_and_initialize_db(dbfile, TARGET_VERSION)
    return db

def dump_db(db):
    # to let _iterdump work, we need to restore the original row factory
    orig = db.row_factory
    try:
        db.row_factory = sqlite3.Row
        return "".join(db.iterdump())
    finally:
        db.row_factory = orig



from twisted.internet.protocol import (
    ClientFactory,
    Protocol,
)
from twisted.test import iosim
from zope.interface import (
    Interface,
    Attribute,
    implementer,
)
from ..transit_server import (
    Transit,
    TransitConnection,
)
from twisted.internet.protocol import ServerFactory
from ..usage import create_usage_tracker


class IRelayTestClient(Interface):
    """
    The client interface used by tests.
    """

    connected = Attribute("True if we are currently connected else False")

    def send(data):
        """
        Send some bytes.
        :param bytes data: the data to send
        """

    def disconnect():
        """
        Terminate the connection.
        """

    def get_received_data():
        """
        :returns: all the bytes received from the server on this
        connection.
        """

    def reset_data():
        """
        Erase any received data to this point.
        """


class ServerBase:
    log_requests = False

    def setUp(self):
        self._pumps = []
        self._lp = None
        if self.log_requests:
            blur_usage = None
        else:
            blur_usage = 60.0
        self._setup_relay(blur_usage=blur_usage)

    def flush(self):
        did_work = False
        for pump in self._pumps:
            did_work = pump.flush() or did_work
        if did_work:
            self.flush()

    def _setup_relay(self, blur_usage=None, log_file=None, usage_db=None):
        usage = create_usage_tracker(
            blur_usage=blur_usage,
            log_file=log_file,
            usage_db=usage_db,
        )
        self._transit_server = Transit(usage, lambda: 123456789.0)

    def new_protocol(self):
        """
        This should be overridden by derived test-case classes to decide
        if they want a TCP or WebSockets protocol.
        """
        raise NotImplementedError()

    def new_protocol_tcp(self):
        """
        Create a new client protocol connected to the server.
        :returns: a IRelayTestClient implementation
        """
        server_factory = ServerFactory()
        server_factory.protocol = TransitConnection
        server_factory.transit = self._transit_server
        server_factory.log_requests = self.log_requests
        server_protocol = server_factory.buildProtocol(('127.0.0.1', 0))

        @implementer(IRelayTestClient)
        class TransitClientProtocolTcp(Protocol):
            """
            Speak the transit client protocol used by the tests over TCP
            """
            _received = b""
            connected = False

            # override Protocol callbacks

            def connectionMade(self):
                self.connected = True
                return Protocol.connectionMade(self)

            def connectionLost(self, reason):
                self.connected = False
                return Protocol.connectionLost(self, reason)

            def dataReceived(self, data):
                self._received = self._received + data

            # IRelayTestClient

            def send(self, data):
                self.transport.write(data)

            def disconnect(self):
                self.transport.loseConnection()

            def reset_received_data(self):
                self._received = b""

            def get_received_data(self):
                return self._received

        client_factory = ClientFactory()
        client_factory.protocol = TransitClientProtocolTcp
        client_protocol = client_factory.buildProtocol(('127.0.0.1', 31337))

        pump = iosim.connect(
            server_protocol,
            iosim.makeFakeServer(server_protocol),
            client_protocol,
            iosim.makeFakeClient(client_protocol),
        )
        pump.flush()
        self._pumps.append(pump)
        return client_protocol

    def tearDown(self):
        if self._lp:
            return self._lp.stopListening()



from binascii import hexlify
from twisted.trial import unittest
from twisted.test import iosim
from autobahn.twisted.websocket import (
    WebSocketServerFactory,
    WebSocketClientFactory,
    WebSocketClientProtocol,
)
from autobahn.twisted.testing import (
    create_pumper,
    MemoryReactorClockResolver,
)
from autobahn.exception import Disconnected
from zope.interface import implementer
from .common import (
    ServerBase,
    IRelayTestClient,
)
from ..usage import (
    MemoryUsageRecorder,
    blur_size,
)
from ..transit_server import (
    WebSocketTransitConnection,
    TransitServerState,
)


def handshake(token, side=None):
    hs = b"please relay " + hexlify(token)
    if side is not None:
        hs += b" for side " + hexlify(side)
    hs += b"\n"
    return hs

class _Transit:
    def count(self):
        return sum([
            len(potentials)
            for potentials
            in self._transit_server.pending_requests._requests.values()
        ])

    def test_blur_size(self):
        self.failUnlessEqual(blur_size(0), 0)
        self.failUnlessEqual(blur_size(1), 10e3)
        self.failUnlessEqual(blur_size(10e3), 10e3)
        self.failUnlessEqual(blur_size(10e3+1), 20e3)
        self.failUnlessEqual(blur_size(15e3), 20e3)
        self.failUnlessEqual(blur_size(20e3), 20e3)
        self.failUnlessEqual(blur_size(1e6), 1e6)
        self.failUnlessEqual(blur_size(1e6+1), 2e6)
        self.failUnlessEqual(blur_size(1.5e6), 2e6)
        self.failUnlessEqual(blur_size(2e6), 2e6)
        self.failUnlessEqual(blur_size(900e6), 900e6)
        self.failUnlessEqual(blur_size(1000e6), 1000e6)
        self.failUnlessEqual(blur_size(1050e6), 1100e6)
        self.failUnlessEqual(blur_size(1100e6), 1100e6)
        self.failUnlessEqual(blur_size(1150e6), 1200e6)

    def test_register(self):
        p1 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8

        p1.send(handshake(token1, side1))
        self.flush()
        self.assertEqual(self.count(), 1)

        p1.disconnect()
        self.flush()
        self.assertEqual(self.count(), 0)

        # the token should be removed too
        self.assertEqual(len(self._transit_server.pending_requests._requests), 0)

    def test_both_unsided(self):
        p1 = self.new_protocol()
        p2 = self.new_protocol()

        token1 = b"\x00"*32
        p1.send(handshake(token1, side=None))
        self.flush()
        p2.send(handshake(token1, side=None))
        self.flush()

        # a correct handshake yields an ack, after which we can send
        exp = b"ok\n"
        self.assertEqual(p1.get_received_data(), exp)
        self.assertEqual(p2.get_received_data(), exp)

        p1.reset_received_data()
        p2.reset_received_data()

        s1 = b"data1"
        p1.send(s1)
        self.flush()
        self.assertEqual(p2.get_received_data(), s1)

        p1.disconnect()
        self.flush()

    def test_sided_unsided(self):
        p1 = self.new_protocol()
        p2 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        p1.send(handshake(token1, side=side1))
        self.flush()
        p2.send(handshake(token1, side=None))
        self.flush()

        # a correct handshake yields an ack, after which we can send
        exp = b"ok\n"
        self.assertEqual(p1.get_received_data(), exp)
        self.assertEqual(p2.get_received_data(), exp)

        p1.reset_received_data()
        p2.reset_received_data()

        # all data they sent after the handshake should be given to us
        s1 = b"data1"
        p1.send(s1)
        self.flush()
        self.assertEqual(p2.get_received_data(), s1)

        p1.disconnect()
        self.flush()

    def test_unsided_sided(self):
        p1 = self.new_protocol()
        p2 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        p1.send(handshake(token1, side=None))
        p2.send(handshake(token1, side=side1))
        self.flush()

        # a correct handshake yields an ack, after which we can send
        exp = b"ok\n"
        self.assertEqual(p1.get_received_data(), exp)
        self.assertEqual(p2.get_received_data(), exp)

        p1.reset_received_data()
        p2.reset_received_data()

        # all data they sent after the handshake should be given to us
        s1 = b"data1"
        p1.send(s1)
        self.flush()
        self.assertEqual(p2.get_received_data(), s1)

        p1.disconnect()
        p2.disconnect()

    def test_both_sided(self):
        p1 = self.new_protocol()
        p2 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        side2 = b"\x02"*8
        p1.send(handshake(token1, side=side1))
        self.flush()
        p2.send(handshake(token1, side=side2))
        self.flush()

        # a correct handshake yields an ack, after which we can send
        exp = b"ok\n"
        self.assertEqual(p1.get_received_data(), exp)
        self.assertEqual(p2.get_received_data(), exp)

        p1.reset_received_data()
        p2.reset_received_data()

        # all data they sent after the handshake should be given to us
        s1 = b"data1"
        p1.send(s1)
        self.flush()
        self.assertEqual(p2.get_received_data(), s1)

        p1.disconnect()
        p2.disconnect()

    def test_ignore_same_side(self):
        p1 = self.new_protocol()
        p2 = self.new_protocol()
        p3 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8

        p1.send(handshake(token1, side=side1))
        self.flush()
        self.assertEqual(self.count(), 1)

        p2.send(handshake(token1, side=side1))
        self.flush()
        self.flush()
        self.assertEqual(self.count(), 2) # same-side connections don't match

        # when the second side arrives, the spare first connection should be
        # closed
        side2 = b"\x02"*8
        p3.send(handshake(token1, side=side2))
        self.flush()
        self.assertEqual(self.count(), 0)
        self.assertEqual(len(self._transit_server.pending_requests._requests), 0)
        self.assertEqual(len(self._transit_server.active_connections._connections), 2)
        # That will trigger a disconnect on exactly one of (p1 or p2).
        # The other connection should still be connected
        self.assertEqual(sum([int(t.connected) for t in [p1, p2]]), 1)

        p1.disconnect()
        p2.disconnect()
        p3.disconnect()

    def test_bad_handshake_old(self):
        p1 = self.new_protocol()

        token1 = b"\x00"*32
        p1.send(b"please DELAY " + hexlify(token1) + b"\n")
        self.flush()

        exp = b"bad handshake\n"
        self.assertEqual(p1.get_received_data(), exp)
        p1.disconnect()

    def test_bad_handshake_old_slow(self):
        p1 = self.new_protocol()

        p1.send(b"please DELAY ")
        self.flush()
        # As in test_impatience_new_slow, the current state machine has code
        # that can only be reached if we insert a stall here, so dataReceived
        # gets called twice. Hopefully we can delete this test once
        # dataReceived is refactored to remove that state.

        token1 = b"\x00"*32
        # the server waits for the exact number of bytes in the expected
        # handshake message. to trigger "bad handshake", we must match.
        p1.send(hexlify(token1) + b"\n")
        self.flush()

        exp = b"bad handshake\n"
        self.assertEqual(p1.get_received_data(), exp)

        p1.disconnect()

    def test_bad_handshake_new(self):
        p1 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        # the server waits for the exact number of bytes in the expected
        # handshake message. to trigger "bad handshake", we must match.
        p1.send(b"please DELAY " + hexlify(token1) +
                b" for side " + hexlify(side1) + b"\n")
        self.flush()

        exp = b"bad handshake\n"
        self.assertEqual(p1.get_received_data(), exp)

        p1.disconnect()

    def test_binary_handshake(self):
        p1 = self.new_protocol()

        binary_bad_handshake = b"\x00\x01\xe0\x0f\n\xff"
        # the embedded \n makes the server trigger early, before the full
        # expected handshake length has arrived. A non-wormhole client
        # writing non-ascii junk to the transit port used to trigger a
        # UnicodeDecodeError when it tried to coerce the incoming handshake
        # to unicode, due to the ("\n" in buf) check. This was fixed to use
        # (b"\n" in buf). This exercises the old failure.
        p1.send(binary_bad_handshake)
        self.flush()

        exp = b"bad handshake\n"
        self.assertEqual(p1.get_received_data(), exp)

        p1.disconnect()

    def test_impatience_old(self):
        p1 = self.new_protocol()

        token1 = b"\x00"*32
        # sending too many bytes is impatience.
        p1.send(b"please relay " + hexlify(token1))
        p1.send(b"\nNOWNOWNOW")
        self.flush()

        exp = b"impatient\n"
        self.assertEqual(p1.get_received_data(), exp)

        p1.disconnect()

    def test_impatience_new(self):
        p1 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        # sending too many bytes is impatience.
        p1.send(b"please relay " + hexlify(token1) +
                b" for side " + hexlify(side1))
        p1.send(b"\nNOWNOWNOW")
        self.flush()

        exp = b"impatient\n"
        self.assertEqual(p1.get_received_data(), exp)

        p1.disconnect()

    def test_impatience_new_slow(self):
        p1 = self.new_protocol()
        # For full coverage, we need dataReceived to see a particular framing
        # of these two pieces of data, and ITCPTransport doesn't have flush()
        # (which probably wouldn't work anyways). For now, force a 100ms
        # stall between the two writes. I tried setTcpNoDelay(True) but it
        # didn't seem to help without the stall. The long-term fix is to
        # rewrite dataReceived() to remove the multiple "impatient"
        # codepaths, deleting the particular clause that this test exercises,
        # then remove this test.

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        # sending too many bytes is impatience.
        p1.send(b"please relay " + hexlify(token1) +
                b" for side " + hexlify(side1) + b"\n")
        self.flush()

        p1.send(b"NOWNOWNOW")
        self.flush()

        exp = b"impatient\n"
        self.assertEqual(p1.get_received_data(), exp)

        p1.disconnect()

    def test_short_handshake(self):
        p1 = self.new_protocol()
        # hang up before sending a complete handshake
        p1.send(b"short")
        self.flush()
        p1.disconnect()

    def test_empty_handshake(self):
        p1 = self.new_protocol()
        # hang up before sending anything
        p1.disconnect()


class TransitWithLogs(_Transit, ServerBase, unittest.TestCase):
    log_requests = True

    def new_protocol(self):
        return self.new_protocol_tcp()


class TransitWithoutLogs(_Transit, ServerBase, unittest.TestCase):
    log_requests = False

    def new_protocol(self):
        return self.new_protocol_tcp()


def _new_protocol_ws(transit_server, log_requests):
    """
    Internal helper for test-suites that need to provide WebSocket
    client/server pairs.

    :returns: a 2-tuple: (iosim.IOPump, protocol)
    """
    ws_factory = WebSocketServerFactory("ws://localhost:4002")
    ws_factory.protocol = WebSocketTransitConnection
    ws_factory.transit = transit_server
    ws_factory.log_requests = log_requests
    ws_protocol = ws_factory.buildProtocol(('127.0.0.1', 0))

    @implementer(IRelayTestClient)
    class TransitWebSocketClientProtocol(WebSocketClientProtocol):
        _received = b""
        connected = False

        def connectionMade(self):
            self.connected = True
            return super(TransitWebSocketClientProtocol, self).connectionMade()

        def connectionLost(self, reason):
            self.connected = False
            return super(TransitWebSocketClientProtocol, self).connectionLost(reason)

        def onMessage(self, data, isBinary):
            self._received = self._received + data

        def send(self, data):
            self.sendMessage(data, True)

        def get_received_data(self):
            return self._received

        def reset_received_data(self):
            self._received = b""

        def disconnect(self):
            self.sendClose(1000, True)

    client_factory = WebSocketClientFactory()
    client_factory.protocol = TransitWebSocketClientProtocol
    client_protocol = client_factory.buildProtocol(('127.0.0.1', 31337))
    client_protocol.disconnect = client_protocol.dropConnection

    pump = iosim.connect(
        ws_protocol,
        iosim.makeFakeServer(ws_protocol),
        client_protocol,
        iosim.makeFakeClient(client_protocol),
    )
    return pump, client_protocol



class TransitWebSockets(_Transit, ServerBase, unittest.TestCase):

    def new_protocol(self):
        return self.new_protocol_ws()

    def new_protocol_ws(self):
        pump, proto = _new_protocol_ws(self._transit_server, self.log_requests)
        self._pumps.append(pump)
        return proto

    def test_websocket_to_tcp(self):
        """
        One client is WebSocket and one is TCP
        """
        p1 = self.new_protocol_ws()
        p2 = self.new_protocol_tcp()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        side2 = b"\x02"*8
        p1.send(handshake(token1, side=side1))
        self.flush()
        p2.send(handshake(token1, side=side2))
        self.flush()

        # a correct handshake yields an ack, after which we can send
        exp = b"ok\n"
        self.assertEqual(p1.get_received_data(), exp)
        self.assertEqual(p2.get_received_data(), exp)

        p1.reset_received_data()
        p2.reset_received_data()

        # all data they sent after the handshake should be given to us
        s1 = b"data1"
        p1.send(s1)
        self.flush()
        self.assertEqual(p2.get_received_data(), s1)

        p1.disconnect()
        p2.disconnect()
        self.flush()

    def test_bad_handshake_old_slow(self):
        """
        This test only makes sense for TCP
        """

    def test_send_closed_partner(self):
        """
        Sending data to a closed partner causes an error that propogates
        to the sender.
        """
        p1 = self.new_protocol()
        p2 = self.new_protocol()

        # set up a successful connection
        token = b"a" * 32
        p1.send(handshake(token))
        p2.send(handshake(token))
        self.flush()

        # p2 loses connection, then p1 sends a message
        p2.transport.loseConnection()
        self.flush()

        # at this point, p1 learns that p2 is disconnected (because it
        # tried to relay "a message" but failed)

        # try to send more (our partner p2 is gone now though so it
        # should be an immediate error)
        with self.assertRaises(Disconnected):
            p1.send(b"more message")
            self.flush()


class Usage(ServerBase, unittest.TestCase):
    log_requests = True

    def setUp(self):
        super(Usage, self).setUp()
        self._usage = MemoryUsageRecorder()
        self._transit_server.usage.add_backend(self._usage)

    def new_protocol(self):
        return self.new_protocol_tcp()

    def test_empty(self):
        p1 = self.new_protocol()
        # hang up before sending anything
        p1.disconnect()
        self.flush()

        # that will log the "empty" usage event
        self.assertEqual(len(self._usage.events), 1, self._usage)
        self.assertEqual(self._usage.events[0]["mood"], "empty", self._usage)

    def test_short(self):
        # Note: this test only runs on TCP clients because WebSockets
        # already does framing (so it's either "a bad handshake" or
        # there's no handshake at all yet .. you can't have a "short"
        # one).
        p1 = self.new_protocol()
        # hang up before sending a complete handshake
        p1.send(b"short")
        p1.disconnect()
        self.flush()

        # that will log the "empty" usage event
        self.assertEqual(len(self._usage.events), 1, self._usage)
        self.assertEqual("empty", self._usage.events[0]["mood"])

    def test_errory(self):
        p1 = self.new_protocol()

        p1.send(b"this is a very bad handshake\n")
        self.flush()
        # that will log the "errory" usage event, then drop the connection
        p1.disconnect()
        self.assertEqual(len(self._usage.events), 1, self._usage)
        self.assertEqual(self._usage.events[0]["mood"], "errory", self._usage)

    def test_lonely(self):
        p1 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        p1.send(handshake(token1, side=side1))
        self.flush()
        # now we disconnect before the peer connects
        p1.disconnect()
        self.flush()

        self.assertEqual(len(self._usage.events), 1, self._usage)
        self.assertEqual(self._usage.events[0]["mood"], "lonely", self._usage)
        self.assertIdentical(self._usage.events[0]["waiting_time"], None)

    def test_one_happy_one_jilted(self):
        p1 = self.new_protocol()
        p2 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        side2 = b"\x02"*8
        p1.send(handshake(token1, side=side1))
        self.flush()
        p2.send(handshake(token1, side=side2))
        self.flush()

        self.assertEqual(self._usage.events, []) # no events yet

        p1.send(b"\x00" * 13)
        self.flush()
        p2.send(b"\xff" * 7)
        self.flush()

        p1.disconnect()
        self.flush()

        self.assertEqual(len(self._usage.events), 1, self._usage)
        self.assertEqual(self._usage.events[0]["mood"], "happy", self._usage)
        self.assertEqual(self._usage.events[0]["total_bytes"], 20)
        self.assertNotIdentical(self._usage.events[0]["waiting_time"], None)

    def test_redundant(self):
        p1a = self.new_protocol()
        p1b = self.new_protocol()
        p1c = self.new_protocol()
        p2 = self.new_protocol()

        token1 = b"\x00"*32
        side1 = b"\x01"*8
        side2 = b"\x02"*8
        p1a.send(handshake(token1, side=side1))
        self.flush()
        p1b.send(handshake(token1, side=side1))
        self.flush()

        # connect and disconnect a third client (for side1) to exercise the
        # code that removes a pending connection without removing the entire
        # token
        p1c.send(handshake(token1, side=side1))
        p1c.disconnect()
        self.flush()

        self.assertEqual(len(self._usage.events), 1, self._usage)
        self.assertEqual(self._usage.events[0]["mood"], "lonely")

        p2.send(handshake(token1, side=side2))
        self.flush()
        self.assertEqual(len(self._transit_server.pending_requests._requests), 0)
        self.assertEqual(len(self._usage.events), 2, self._usage)
        self.assertEqual(self._usage.events[1]["mood"], "redundant")

        # one of the these is unecessary, but probably harmless
        p1a.disconnect()
        p1b.disconnect()
        self.flush()
        self.assertEqual(len(self._usage.events), 3, self._usage)
        self.assertEqual(self._usage.events[2]["mood"], "happy")


class UsageWebSockets(Usage):
    """
    All the tests of 'Usage' except with a WebSocket (instead of TCP)
    transport.

    This overrides ServerBase.new_protocol to achieve this. It might
    be nicer to parametrize these tests in a way that doesn't use
    inheritance .. but all the support etc classes are set up that way
    already.
    """

    def setUp(self):
        super(UsageWebSockets, self).setUp()
        self._pump = create_pumper()
        self._reactor = MemoryReactorClockResolver()
        return self._pump.start()

    def tearDown(self):
        return self._pump.stop()

    def new_protocol(self):
        return self.new_protocol_ws()

    def new_protocol_ws(self):
        pump, proto = _new_protocol_ws(self._transit_server, self.log_requests)
        self._pumps.append(pump)
        return proto

    def test_short(self):
        """
        This test essentially just tests the framing of the line-oriented
        TCP protocol; it doesnt' make sense for the WebSockets case
        because WS handles frameing: you either sent a 'bad handshake'
        because it is semantically invalid or no handshake (yet).
        """

    def test_send_non_binary_message(self):
        """
        A non-binary WebSocket message is an error
        """
        ws_factory = WebSocketServerFactory("ws://localhost:4002")
        ws_factory.protocol = WebSocketTransitConnection
        ws_protocol = ws_factory.buildProtocol(('127.0.0.1', 0))
        with self.assertRaises(ValueError):
            ws_protocol.onMessage(u"foo", isBinary=False)


class State(unittest.TestCase):
    """
    Tests related to server_state.TransitServerState
    """

    def setUp(self):
        self.state = TransitServerState(None, None)

    def test_empty_token(self):
        self.assertEqual(
            "-",
            self.state.get_token(),
        )



try:
    # 'resource' is unix-only
    from resource import getrlimit, setrlimit, RLIMIT_NOFILE
except ImportError: # pragma: nocover
    getrlimit, setrlimit, RLIMIT_NOFILE = None, None, None # pragma: nocover
from twisted.python import log

def increase_rlimits():
    if getrlimit is None:
        log.msg("unable to import 'resource', leaving rlimit alone")
        return
    soft, hard = getrlimit(RLIMIT_NOFILE)
    if soft >= 10000:
        log.msg("RLIMIT_NOFILE.soft was %d, leaving it alone" % soft)
        return
    # OS-X defaults to soft=7168, and reports a huge number for 'hard',
    # but won't accept anything more than soft=10240, so we can't just
    # set soft=hard. Linux returns (1024, 1048576) and is fine with
    # soft=hard. Cygwin is reported to return (256,-1) and accepts up to
    # soft=3200. So we try multiple values until something works.
    for newlimit in [hard, 10000, 3200, 1024]:
        log.msg("changing RLIMIT_NOFILE from (%s,%s) to (%s,%s)" %
                (soft, hard, newlimit, hard))
        try:
            setrlimit(RLIMIT_NOFILE, (newlimit, hard))
            log.msg("setrlimit successful")
            return
        except ValueError as e:
            log.msg("error during setrlimit: %s" % e)
            continue
        except:
            log.msg("other error during setrlimit, leaving it alone")
            log.err()
            return
    log.msg("unable to change rlimit, leaving it alone")



from twisted.trial import unittest
from unittest import mock
from twisted.application.service import MultiService
from autobahn.twisted.websocket import WebSocketServerFactory
from .. import server_tap

class Service(unittest.TestCase):
    def test_defaults(self):
        o = server_tap.Options()
        o.parseOptions([])
        with mock.patch("wormhole_transit_relay.server_tap.create_usage_tracker") as t:
            s = server_tap.makeService(o)
        self.assertEqual(t.mock_calls,
                         [mock.call(blur_usage=None,
                                    log_file=None, usage_db=None)])
        self.assertIsInstance(s, MultiService)

    def test_blur(self):
        o = server_tap.Options()
        o.parseOptions(["--blur-usage=60"])
        with mock.patch("wormhole_transit_relay.server_tap.create_usage_tracker") as t:
            server_tap.makeService(o)
        self.assertEqual(t.mock_calls,
                         [mock.call(blur_usage=60,
                                    log_file=None, usage_db=None)])

    def test_log_fd(self):
        o = server_tap.Options()
        o.parseOptions(["--log-fd=99"])
        fd = object()
        with mock.patch("wormhole_transit_relay.server_tap.create_usage_tracker") as t:
            with mock.patch("wormhole_transit_relay.server_tap.os.fdopen",
                            return_value=fd) as f:
                server_tap.makeService(o)
        self.assertEqual(f.mock_calls, [mock.call(99, "w")])
        self.assertEqual(t.mock_calls,
                         [mock.call(blur_usage=None,
                                    log_file=fd, usage_db=None)])

    def test_websocket(self):
        """
        A websocket factory is created when passing --websocket
        """
        o = server_tap.Options()
        o.parseOptions(["--websocket=tcp:4004"])
        services = server_tap.makeService(o)
        self.assertTrue(
            any(
                isinstance(s.factory, WebSocketServerFactory)
                for s in services.services
            )
        )

    def test_websocket_explicit_url(self):
        """
        A websocket factory is created with --websocket and
        --websocket-url
        """
        o = server_tap.Options()
        o.parseOptions([
            "--websocket=tcp:4004",
            "--websocket-url=ws://example.com:4004",
        ])
        services = server_tap.makeService(o)
        self.assertTrue(
            any(
                isinstance(s.factory, WebSocketServerFactory)
                for s in services.services
            )
        )



from io import (
    StringIO,
)
import sys
import shutil

from twisted.trial import unittest
from twisted.internet.interfaces import (
    IPullProducer,
)
from twisted.internet.protocol import (
    ProcessProtocol,
)
from twisted.internet.defer import (
    inlineCallbacks,
    Deferred,
)
from autobahn.twisted.websocket import (
    WebSocketClientProtocol,
    create_client_agent,
)
from zope.interface import implementer


class _CollectOutputProtocol(ProcessProtocol):
    """
    Internal helper. Collects all output (stdout + stderr) into
    self.output, and callback's on done with all of it after the
    process exits (for any reason).
    """
    def __init__(self):
        self.done = Deferred()
        self.running = Deferred()
        self.output = StringIO()

    def processEnded(self, reason):
        if not self.done.called:
            self.done.callback(self.output.getvalue())

    def outReceived(self, data):
        print(data.decode(), end="", flush=True)
        self.output.write(data.decode(sys.getfilesystemencoding()))
        if not self.running.called:
            if "on 8088" in self.output.getvalue():
                self.running.callback(None)

    def errReceived(self, data):
        print("ERR: {}".format(data.decode(sys.getfilesystemencoding())))
        self.output.write(data.decode(sys.getfilesystemencoding()))


def run_transit(reactor, proto, tcp_port=None, websocket_port=None):
    exe = shutil.which("twistd")
    args = [
        exe, "-n", "transitrelay",
    ]
    if tcp_port is not None:
        args.append("--port")
        args.append(tcp_port)
    if websocket_port is not None:
        args.append("--websocket")
        args.append(websocket_port)
    proc = reactor.spawnProcess(proto, exe, args)
    return proc



class Sender(WebSocketClientProtocol):
    """
    """

    def __init__(self, *args, **kw):
        WebSocketClientProtocol.__init__(self, *args, **kw)
        self.done = Deferred()
        self.got_ok = Deferred()

    def onMessage(self, payload, is_binary):
        print("onMessage")
        if not self.got_ok.called:
            if payload == b"ok\n":
                self.got_ok.callback(None)
        print("send: {}".format(payload.decode("utf8")))

    def onClose(self, clean, code, reason):
        print(f"close: {clean} {code} {reason}")
        self.done.callback(None)


class Receiver(WebSocketClientProtocol):
    """
    """

    def __init__(self, *args, **kw):
        WebSocketClientProtocol.__init__(self, *args, **kw)
        self.done = Deferred()
        self.first_message = Deferred()
        self.received = 0

    def onMessage(self, payload, is_binary):
        print("recv: {}".format(len(payload)))
        self.received += len(payload)
        if not self.first_message.called:
            self.first_message.callback(None)

    def onClose(self, clean, code, reason):
        print(f"close: {clean} {code} {reason}")
        self.done.callback(None)


class TransitWebSockets(unittest.TestCase):
    """
    Integration-style tests of the transit WebSocket relay, using the
    real reactor (and running transit as a subprocess).
    """

    @inlineCallbacks
    def test_buffer_fills(self):
        """
        A running transit relay stops accepting incoming data at a
        reasonable amount if the peer isn't reading. This test defines
        that as 'less than 100MiB' although in practice Twisted seems
        to stop before 10MiB.
        """
        from twisted.internet import reactor
        transit_proto = _CollectOutputProtocol()
        transit_proc = run_transit(reactor, transit_proto, websocket_port="tcp:8088")

        def cleanup_process():
            transit_proc.signalProcess("HUP")
            return transit_proto.done
        self.addCleanup(cleanup_process)

        yield transit_proto.running
        print("Transit running")

        agent = create_client_agent(reactor)
        side_a = yield agent.open("ws://localhost:8088", {}, lambda: Sender())
        side_b = yield agent.open("ws://localhost:8088", {}, lambda: Receiver())

        side_a.sendMessage(b"please relay aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa for side aaaaaaaaaaaaaaaa", True)
        side_b.sendMessage(b"please relay aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa for side bbbbbbbbbbbbbbbb", True)

        yield side_a.got_ok
        yield side_b.first_message

        # remove side_b's filedescriptor from the reactor .. this
        # means it will not read any more data
        reactor.removeReader(side_b.transport)

        # attempt to send up to 100MiB through side_a .. we should get
        # backpressure before that works which only manifests itself
        # as this producer not being asked to produce more
        max_data = 1024*1024*100 # 100MiB

        @implementer(IPullProducer)
        class ProduceMessages:
            def __init__(self, ws, on_produce):
                self._ws = ws
                self._sent = 0
                self._max = max_data
                self._on_produce = on_produce

            def resumeProducing(self):
                self._on_produce()
                if self._sent >= self._max:
                    self._ws.sendClose()
                    return
                data = b"a" * 1024*1024
                self._ws.sendMessage(data, True)
                self._sent += len(data)
                print("sent {}, total {}".format(len(data), self._sent))

        # our only signal is, "did our producer get asked to produce
        # more data" which it should do periodically. We want to stop
        # if we haven't seen a new data request for a while -- defined
        # as "more than 5 seconds".

        done = Deferred()
        last_produce = None
        timeout = 2  # seconds

        def asked_for_data():
            nonlocal last_produce
            last_produce = reactor.seconds()

        data = ProduceMessages(side_a, asked_for_data)
        side_a.transport.registerProducer(data, False)
        data.resumeProducing()

        def check_if_done():
            if last_produce is not None:
                if reactor.seconds() - last_produce > timeout:
                    done.callback(None)
                    return
            # recursive call to ourselves to check again soon
            reactor.callLater(.1, check_if_done)
        check_if_done()

        yield done

        mib = 1024*1024.0
        print("Sent {}MiB of {}MiB before backpressure".format(data._sent / mib, max_data / mib))
        self.assertTrue(data._sent < max_data, "Too much data sent")

        side_a.sendClose()
        side_b.sendClose()
        yield side_a.done
        yield side_b.done



import time
import json

from twisted.python import log
from zope.interface import (
    implementer,
    Interface,
)


def create_usage_tracker(blur_usage, log_file, usage_db):
    """
    :param int blur_usage: see UsageTracker

    :param log_file: None or a file-like object to write JSON-encoded
        lines of usage information to.

    :param usage_db: None or an sqlite3 database connection

    :returns: a new UsageTracker instance configured with backends.
    """
    tracker = UsageTracker(blur_usage)
    if usage_db:
        tracker.add_backend(DatabaseUsageRecorder(usage_db))
    if log_file:
        tracker.add_backend(LogFileUsageRecorder(log_file))
    return tracker


class IUsageWriter(Interface):
    """
    Records actual usage statistics in some way
    """

    def record_usage(started=None, total_time=None, waiting_time=None, total_bytes=None, mood=None):
        """
        :param int started: timestemp when this connection began

        :param float total_time: total seconds this connection lasted

        :param float waiting_time: None or the total seconds one side
            waited for the other

        :param int total_bytes: the total bytes sent. In case the
            connection was concluded successfully, only one side will
            record the total bytes (but count both).

        :param str mood: the 'mood' of the connection
        """


@implementer(IUsageWriter)
class MemoryUsageRecorder:
    """
    Remebers usage records in memory.
    """

    def __init__(self):
        self.events = []

    def record_usage(self, started=None, total_time=None, waiting_time=None, total_bytes=None, mood=None):
        """
        IUsageWriter.
        """
        data = {
            "started": started,
            "total_time": total_time,
            "waiting_time": waiting_time,
            "total_bytes": total_bytes,
            "mood": mood,
        }
        self.events.append(data)


@implementer(IUsageWriter)
class LogFileUsageRecorder:
    """
    Writes usage records to a file. The records are written in JSON,
    one record per line.
    """

    def __init__(self, writable_file):
        self._file = writable_file

    def record_usage(self, started=None, total_time=None, waiting_time=None, total_bytes=None, mood=None):
        """
        IUsageWriter.
        """
        data = {
            "started": started,
            "total_time": total_time,
            "waiting_time": waiting_time,
            "total_bytes": total_bytes,
            "mood": mood,
        }
        self._file.write(json.dumps(data) + "\n")
        self._file.flush()


@implementer(IUsageWriter)
class DatabaseUsageRecorder:
    """
    Write usage records into a database
    """

    def __init__(self, db):
        self._db = db

    def record_usage(self, started=None, total_time=None, waiting_time=None, total_bytes=None, mood=None):
        """
        IUsageWriter.
        """
        self._db.execute(
            "INSERT INTO `usage`"
            " (`started`, `total_time`, `waiting_time`,"
            "  `total_bytes`, `result`)"
            " VALUES (?,?,?,?,?)",
            (started, total_time, waiting_time, total_bytes, mood)
        )
        # original code did "self._update_stats()" here, thus causing
        # "global" stats update on every connection update .. should
        # we repeat this behavior, or really only record every
        # 60-seconds with the timer?
        self._db.commit()


class UsageTracker(object):
    """
    Tracks usage statistics of connections
    """

    def __init__(self, blur_usage):
        """
        :param int blur_usage: None or the number of seconds to use as a
            window around which to blur time statistics (e.g. "60" means times
            will be rounded to 1 minute intervals). When blur_usage is
            non-zero, sizes will also be rounded into buckets of "one
            megabyte", "one gigabyte" or "lots"
        """
        self._backends = set()
        self._blur_usage = blur_usage
        if blur_usage:
            log.msg("blurring access times to %d seconds" % self._blur_usage)
        else:
            log.msg("not blurring access times")

    def add_backend(self, backend):
        """
        Add a new backend.

        :param IUsageWriter backend: the backend to add
        """
        self._backends.add(backend)

    def record(self, started, buddy_started, result, bytes_sent, buddy_bytes):
        """
        :param int started: timestamp when our connection started

        :param int buddy_started: None, or the timestamp when our
            partner's connection started (will be None if we don't yet
            have a partner).

        :param str result: a label for the result of the connection
            (one of the "moods").

        :param int bytes_sent: number of bytes we sent

        :param int buddy_bytes: number of bytes our partner sent
        """
        # ideally self._reactor.seconds() or similar, but ..
        finished = time.time()
        if buddy_started is not None:
            starts = [started, buddy_started]
            total_time = finished - min(starts)
            waiting_time = max(starts) - min(starts)
            total_bytes = bytes_sent + buddy_bytes
        else:
            total_time = finished - started
            waiting_time = None
            total_bytes = bytes_sent
            # note that "bytes_sent" should always be 0 here, but
            # we're recording what the state-machine remembered in any
            # case

        if self._blur_usage:
            started = self._blur_usage * (started // self._blur_usage)
            total_bytes = blur_size(total_bytes)

        # This is "a dict" instead of "kwargs" because we have to make
        # it into a dict for the log use-case and in-memory/testing
        # use-case anyway so this is less repeats of the names.
        self._notify_backends({
            "started": started,
            "total_time": total_time,
            "waiting_time": waiting_time,
            "total_bytes": total_bytes,
            "mood": result,
        })

    def update_stats(self, rebooted, updated, connected, waiting,
                     incomplete_bytes):
        """
        Update general statistics.
        """
        # in original code, this is only recorded in the database
        # .. perhaps a better way to do this, but ..
        for backend in self._backends:
            if isinstance(backend, DatabaseUsageRecorder):
                backend._db.execute("DELETE FROM `current`")
                backend._db.execute(
                    "INSERT INTO `current`"
                    " (`rebooted`, `updated`, `connected`, `waiting`,"
                    "  `incomplete_bytes`)"
                    " VALUES (?, ?, ?, ?, ?)",
                    (int(rebooted), int(updated), connected, waiting,
                     incomplete_bytes)
                )

    def _notify_backends(self, data):
        """
        Internal helper. Tell every backend we have about a new usage record.
        """
        for backend in self._backends:
            backend.record_usage(**data)


def round_to(size, coarseness):
    return int(coarseness*(1+int((size-1)/coarseness)))


def blur_size(size):
    if size == 0:
        return 0
    if size < 1e6:
        return round_to(size, 10e3)
    if size < 1e9:
        return round_to(size, 1e6)
    return round_to(size, 100e6)



import os
from twisted.python import filepath
from twisted.trial import unittest
from .. import database
from ..database import get_db, TARGET_VERSION, dump_db, DBError

class Get(unittest.TestCase):
    def test_create_default(self):
        db_url = ":memory:"
        db = get_db(db_url)
        rows = db.execute("SELECT * FROM version").fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0]["version"], TARGET_VERSION)

    def test_open_existing_file(self):
        basedir = self.mktemp()
        os.mkdir(basedir)
        fn = os.path.join(basedir, "normal.db")
        db = get_db(fn)
        rows = db.execute("SELECT * FROM version").fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0]["version"], TARGET_VERSION)
        db2 = get_db(fn)
        rows = db2.execute("SELECT * FROM version").fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0]["version"], TARGET_VERSION)

    def test_open_bad_version(self):
        basedir = self.mktemp()
        os.mkdir(basedir)
        fn = os.path.join(basedir, "old.db")
        db = get_db(fn)
        db.execute("UPDATE version SET version=999")
        db.commit()

        with self.assertRaises(DBError) as e:
            get_db(fn)
        self.assertIn("Unable to handle db version 999", str(e.exception))

    def test_open_corrupt(self):
        basedir = self.mktemp()
        os.mkdir(basedir)
        fn = os.path.join(basedir, "corrupt.db")
        with open(fn, "wb") as f:
            f.write(b"I am not a database")
        with self.assertRaises(DBError) as e:
            get_db(fn)
        self.assertIn("not a database", str(e.exception))

    def test_failed_create_allows_subsequent_create(self):
        patch = self.patch(database, "get_schema", lambda version: b"this is a broken schema")
        dbfile = filepath.FilePath(self.mktemp())
        self.assertRaises(Exception, lambda: get_db(dbfile.path))
        patch.restore()
        get_db(dbfile.path)

    def OFF_test_upgrade(self): # disabled until we add a v2 schema
        basedir = self.mktemp()
        os.mkdir(basedir)
        fn = os.path.join(basedir, "upgrade.db")
        self.assertNotEqual(TARGET_VERSION, 2)

        # create an old-version DB in a file
        db = get_db(fn, 2)
        rows = db.execute("SELECT * FROM version").fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0]["version"], 2)
        del db

        # then upgrade the file to the latest version
        dbA = get_db(fn, TARGET_VERSION)
        rows = dbA.execute("SELECT * FROM version").fetchall()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0]["version"], TARGET_VERSION)
        dbA_text = dump_db(dbA)
        del dbA

        # make sure the upgrades got committed to disk
        dbB = get_db(fn, TARGET_VERSION)
        dbB_text = dump_db(dbB)
        del dbB
        self.assertEqual(dbA_text, dbB_text)

        # The upgraded schema should be equivalent to that of a new DB.
        # However a text dump will differ because ALTER TABLE always appends
        # the new column to the end of a table, whereas our schema puts it
        # somewhere in the middle (wherever it fits naturally). Also ALTER
        # TABLE doesn't include comments.
        if False:
            latest_db = get_db(":memory:", TARGET_VERSION)
            latest_text = dump_db(latest_db)
            with open("up.sql","w") as f: f.write(dbA_text)
            with open("new.sql","w") as f: f.write(latest_text)
            # check with "diff -u _trial_temp/up.sql _trial_temp/new.sql"
            self.assertEqual(dbA_text, latest_text)

class Create(unittest.TestCase):
    def test_memory(self):
        db = database.create_db(":memory:")
        latest_text = dump_db(db)
        self.assertIn("CREATE TABLE", latest_text)

    def test_preexisting(self):
        basedir = self.mktemp()
        os.mkdir(basedir)
        fn = os.path.join(basedir, "preexisting.db")
        with open(fn, "w"):
            pass
        with self.assertRaises(database.DBAlreadyExists):
            database.create_db(fn)

    def test_create(self):
        basedir = self.mktemp()
        os.mkdir(basedir)
        fn = os.path.join(basedir, "created.db")
        db = database.create_db(fn)
        latest_text = dump_db(db)
        self.assertIn("CREATE TABLE", latest_text)

class Open(unittest.TestCase):
    def test_open(self):
        basedir = self.mktemp()
        os.mkdir(basedir)
        fn = os.path.join(basedir, "created.db")
        db1 = database.create_db(fn)
        latest_text = dump_db(db1)
        self.assertIn("CREATE TABLE", latest_text)
        db2 = database.open_existing_db(fn)
        self.assertIn("CREATE TABLE", dump_db(db2))

    def test_doesnt_exist(self):
        basedir = self.mktemp()
        os.mkdir(basedir)
        fn = os.path.join(basedir, "created.db")
        with self.assertRaises(database.DBDoesntExist):
            database.open_existing_db(fn)


